@@ -2,7 +2,7 @@
 # Purpose : Generic Cache Factory with various policy factories.
 # Author  : Sam Graham
 # Created : 23 Jun 2008
-# CVS     : $Id: CacheFactory.pm,v 1.5 2008-07-04 21:11:09 illusori Exp $
+# CVS     : $Id: CacheFactory.pm,v 1.9 2008-07-16 10:15:52 illusori Exp $
 ###############################################################################
 
 package Cache::CacheFactory;
@@ -20,14 +20,16 @@ use Cache::CacheFactory::Object;
 
 use base qw/Cache::Cache/;
 
-$Cache::CacheFactory::VERSION =
-    sprintf"%d.%03d", q$Revision: 1.5 $ =~ /: (\d+)\.(\d+)/;
+$Cache::CacheFactory::VERSION = '1.07';
+
+$Cache::CacheFactory::NO_MAX_SIZE = -1;
 
 @Cache::CacheFactory::EXPORT    = qw();
 @Cache::CacheFactory::EXPORT_OK = qw(
     best_available_storage_policy
     best_available_pruning_policy
     best_available_validity_policy
+    $NO_MAX_SIZE
     );
 %Cache::CacheFactory::EXPORT_TAGS = (
     best_available => [ qw(
@@ -88,6 +90,10 @@ sub new
     $self->{ nonwarning_missing_policies } = 1
         if $options{ nonwarning_missing_policies };
 
+    #  Do we deeply clone our data when setting it?
+    $self->{ no_deep_clone } = 1
+        if $options{ no_deep_clone };
+
     #
     #  Grab our policies from the options.
     $self->set_storage_policies(  $options{ storage  } );
@@ -151,7 +157,8 @@ sub set
         }
     }
 
-    $param->{ created_at } = time() unless $param->{ created_at };
+    $param->{ created_at }    = time() unless $param->{ created_at };
+    $param->{ no_deep_clone } = 1      if $self->{ no_deep_clone };
 
     #  Create Cache::CacheFactory::Object instance.
     $object = $self->new_cache_entry_object();
@@ -163,8 +170,41 @@ sub set
         $key, $object, $param );
     $self->foreach_driver( 'pruning',  'set_object_pruning',
         $key, $object, $param );
-    $self->foreach_driver( 'storage',  'set_object',
-        $key, $object, $param );
+    if( $param->{ no_deep_clone } )
+    {
+        #  Since most Cache::Cache's do their own deep cloning
+        #  we try a bit of a hack to try to bypass that.
+        $self->foreach_policy( 'storage',
+            sub
+            {
+                my ( $self, $policy, $storage ) = @_;
+
+                #  Only try this hack on things that subclass behaviour
+                #  we understand.
+                if( $storage->isa( 'Cache::BaseCache' ) )
+                {
+                    my ( $backend );
+
+                    if( $backend = $storage->_get_backend() )
+                    {
+                        $object->set_size( undef );
+                        $object->set_key( undef );
+
+                        $backend->store( $storage->get_namespace(),
+                            $key, $object );
+                        return;
+                    }
+                }
+
+                #  Ok, we couldn't figure out how to do our dirty hack...
+                $storage->set_object( $key, $object, $param );
+            } );
+    }
+    else
+    {
+        $self->foreach_driver( 'storage',  'set_object',
+            $key, $object, $param );
+    }
 
     $self->auto_purge( 'set' );
 }
@@ -227,7 +267,11 @@ sub set_object
     #  Backwards compat with Cache::Object objects.
     unless( $object->isa( 'Cache::CacheFactory::Object' ) )
     {
-        $object = Cache::CacheFactory::Object->new_from_old( $object );
+        my ( $param );
+
+        $param = {};
+        $param->{ no_deep_clone } = 1 if $self->{ no_deep_clone };
+        $object = Cache::CacheFactory::Object->new_from_old( $object, $param );
         #  TODO: compat with expires_at
     }
 
@@ -243,30 +287,30 @@ sub remove
 
 sub Clear
 {
-    my ( $self ) = @_;
+    my ( $self, @args ) = @_;
 
-    $self->foreach_driver( 'storage', 'Clear' );
+    $self->foreach_driver( 'storage', 'Clear', @args );
 }
 
 sub clear
 {
-    my ( $self ) = @_;
+    my ( $self, @args ) = @_;
 
-    $self->foreach_driver( 'storage', 'clear' );
+    $self->foreach_driver( 'storage', 'clear', @args );
 }
 
 sub Purge
 {
-    my ( $self ) = @_;
+    my ( $self, @args ) = @_;
 
-    $self->purge();
+    $self->purge( @args );
 }
 
 sub purge
 {
-    my ( $self ) = @_;
+    my ( $self, @args ) = @_;
 
-    $self->foreach_driver( 'pruning', 'purge', $self );
+    $self->foreach_driver( 'pruning', 'purge', $self, @args );
 }
 
 sub auto_purge
@@ -293,7 +337,7 @@ sub auto_purge
 
 sub Size
 {
-    my ( $self ) = @_;
+    my ( $self, @args ) = @_;
     my ( $size );
 
     $size = 0;
@@ -302,7 +346,7 @@ sub Size
         {
             my ( $self, $policy, $driver ) = @_;
 
-            $size += $driver->Size();
+            $size += $driver->Size( @args );
         } );
 
     return( $size );
@@ -430,6 +474,22 @@ sub get_default_expires_in
     return( $time_pruning->get_default_expires_in() )  if $time_pruning;
 }
 
+sub limit_size
+{
+    my ( $self, $size ) = @_;
+    my ( $size_policy );
+
+    $size_policy = $self->get_policy_driver( 'pruning', 'size' );
+
+    unless( $size_policy )
+    {
+        carp "Cannot limit_size() when no 'size' pruning policy is set.";
+        return;
+    }
+
+    $size_policy->limit_size( $self, $size );
+}
+
 sub set_last_auto_purge
 {
     my ( $self, $last_auto_purge ) = @_;
@@ -829,6 +889,10 @@ See L</"OPTIONS"> below for more details on possible options.
 
 Associates C<$data> with C<$key> in the cache.
 
+A deep copy of C<$data> will automatically be taken if it is a reference,
+you can turn this behaviour off with the cache option C<no_deep_copy>
+detailed in L</"OPTIONS"> below.
+
 C<$expires_in> indicates the time in seconds until this data should be
 erased, or the constant C<$EXPIRES_NOW>, or the constant C<$EXPIRES_NEVER>.
 Defaults to C<$EXPIRES_NEVER>. This variable can also be in the extended
@@ -996,13 +1060,13 @@ Sets or gets the timestamp of the last auto-purge.
 See the documention for C<last_auto_purge> in L</"OPTIONS">
 for further details.
 
-=item $cache->set_auto_purge_on_set( 0 | 1 )
+=item $cache->set_auto_purge_on_set( 0 | 1 );
 
-=item $cache->set_auto_purge_on_get( 0 | 1 )
+=item $cache->set_auto_purge_on_get( 0 | 1 );
 
-=item $boolean = $cache->get_auto_purge_on_set()
+=item $boolean = $cache->get_auto_purge_on_set();
 
-=item $boolean = $cache->get_auto_purge_on_get()
+=item $boolean = $cache->get_auto_purge_on_get();
 
 Turns auto-purging on/off for C<< $cache->set() >> or
 C<< $cache->get() >>, or returns the current state
@@ -1029,6 +1093,15 @@ C<auto_purge_on_get_interval> options.
 
 Look at L</"OPTIONS"> for further details.
 
+=item $cache->limit_size( $size );
+
+Only available if a pruning policy of 'size' has been set,
+this method will allow you to perform a one-off prune of
+the storage policies to C<$size> size or below.
+
+This behaves like the C<limit_size()> method of
+L<Cache::SizeAwareCache>.
+
 =back
 
 =head1 NON-OBJECT-ORIENTATED FUNCTIONS
@@ -1065,6 +1138,24 @@ supply C<:best_available> on the use line to import them.
 
 =back
 
+=head1 CONSTANTS
+
+You can export the following constants:
+
+=over
+
+=item $NO_MAX_SIZE
+
+You can export this with C<< use Cache::CacheFactory qw/$NO_MAX_SIZE/; >>
+and supply it to the C<max_size> option of a 'size' pruning policy.
+
+This value of C<$NO_MAX_SIZE> is compatible with that defined by
+L<Cache::SizeAwareCache>, so you can use either source.
+
+See L<Cache::CacheFactory::Expiry::Size> for further details.
+
+=back
+
 =head1 OPTIONS
 
 The following options may be passed to the C<new()> constructor:
@@ -1244,6 +1335,32 @@ is missing and will instead generate a C<warn>.
 If you also set C<nonwarning_missing_policies> to a true value,
 this C<warn> will also be surpressed.
 
+=item no_deep_clone => 0 | 1
+
+Setting C<no_deep_clone> to a true value will prevent the
+default behaviour of taking a deep clone of the data provided to
+C<< $cache->set() >>.
+
+This can be a performance gain if you don't need to be paranoid
+about the cache sharing references with whatever handed them in,
+or if you want to handle the cloning yourself within your application.
+
+Regretfully C<no_deep_clone> on the cache can only act in an
+advisory capacity to storage policies, they may choose to
+disregard the flag and many of the L<Cache::Cache> modules
+will do just this. (Not unreasonably: they predate
+L<Cache::CacheFactory> considerably.) L<Cache::CacheFactory>
+tries hard to convince them to avoid taking clones but may or
+may not succeed depending on precisely what you're attempting,
+you'll have to suck it and see I'm afraid.
+
+Using this option with a storage policy of 'memory' will provide
+you with similar behaviour to L<Cache::FastMemoryCache>, with
+the exception that, unavoidably, a deep clone is always created
+on C<< $cache->get() >>. If this is undesirable, install
+L<Cache::FastMemoryCache> and use a storage policy of 'fastmemory'
+in conjunction with setting C<no_deep_clone>.
+
 =back
 
 =head1 POLICIES
@@ -1281,6 +1398,13 @@ per-process in-memory caching.
 Implemented using L<Cache::SharedMemoryCache>,
 this provides in-memory caching with the cache shared between processes.
 
+=item fastmemory
+
+Implemented using L<Cache::FastMemoryCache>,
+this provides in-memory caching like the 'memory' policy but with
+all the deep-copies of data stripped out, best used in conjunction
+with the C<no_deep_clone> option set on the cache.
+
 =item null
 
 Implemented using L<Cache::NullCache>, this cache is
@@ -1526,10 +1650,6 @@ version if they prove useful and won't over-complicate matters - for
 now it's best to create a wrapper module that internally creates the
 caches seperately but presents the Cache::Cache API externally.
 
-=item $cache->limit_size( $size ) not implemented for size caches
-
-This is on the future development TODO list.
-
 =back
 
 =head1 SEE ALSO
@@ -1538,7 +1658,8 @@ L<Cache::Cache>, L<Cache::CacheFactory::Object>,
 L<Cache::CacheFactory::Expiry::Base>,
 L<Cache::CacheFactory::Expiry::Time>,
 L<Cache::CacheFactory::Expiry::Size>,
-L<Cache::CacheFactory::Expiry::LastModified>
+L<Cache::CacheFactory::Expiry::LastModified>,
+L<Cache::FastMemoryCache>
 
 =head1 SUPPORT
 
@@ -1583,6 +1704,9 @@ work is done by this module and its subclasses.
 Chris Winters for L<Class::Factory>, saving me the trouble of finding
 out what policy modules are or aren't available.
 
+John Millaway for L<Cache::FastMemoryCache>, which inspired the
+C<no_deep_clone> option.
+
 =head1 COPYRIGHT & LICENSE
 
 Copyright 2008 Sam Graham, all rights reserved.
